use graph_cycles::Cycles;
use serde::Deserialize;

#[derive(Debug, Deserialize)]
pub struct Config {
    pub service: Vec<CService>,

    #[serde(default)]
    pub autostart: Vec<String>
}

#[derive(Debug, Deserialize)]
pub struct CService {
    pub name: String,
    pub command: String,
    pub directory: Option<String>,

    #[serde(default)]
    pub depends_on: Vec<String>
}

impl Config {
    pub fn load() -> Result<Config, String> {
        let data = std::fs::read_to_string("minitd.toml")
            .map_err(|err| format!("Failed to read config file: {err}"))?;
        let config = toml::from_str::<Config>(&data)
            .map_err(|err| format!("Failed to parse config: {err}"))?;

        let mut graph = petgraph::Graph::new();
        let mut node_map = std::collections::HashMap::new();
        for service in &config.service {
            let node = graph.add_node(&service.name);
            node_map.insert(service.name.clone(), node);
        }
        for service in &config.service {
            let my_id = node_map.get(&service.name).unwrap();
            for dep in &service.depends_on {
                let other_id = match node_map.get(dep) {
                    Some(v) => v,
                    None => return Err(format!("Service '{}' depends on unknown service '{}'", service.name, dep))
                };
                graph.add_edge(*my_id, *other_id, ());
            }
        }
        let cycles = graph.cycles();
        if !cycles.is_empty() {
            let mut msg = String::new();
            for cycle in cycles {
                msg += "Dependency cycle: ";
                for node in &cycle {
                    msg += graph.node_weight(*node).unwrap();
                    msg += " -> ";
                }
                msg += graph.node_weight(cycle[0]).unwrap();
                msg += "\n";
            }
            return Err(msg);
        }

        for auto in &config.autostart {
            if config.service.iter().find(|item| item.name == *auto).is_none() {
                return Err(format!("autostart: unkown service '{auto}'"));
            }
        }

        Ok(config)
    }
}
