use std::collections::VecDeque;
use crate::{Data, WorkItem};
use crate::service::ServiceStatus;

pub const EPOLL_SIGFD_ID: u64 = 0;
pub const EPOLL_WEBSERVER_ID: u64 = 1;

const EPOLL_MAX_EVENTS: usize = 16;
static SHUTDOWN_STARTED: std::sync::atomic::AtomicBool = std::sync::atomic::AtomicBool::new(false);

pub fn watch_services(autostart: Vec<String>) {
    let starter = std::thread::spawn(move || starter_thread(autostart));

    let data = Data::get();
    let mut events = [nix::sys::epoll::EpollEvent::empty(); EPOLL_MAX_EVENTS];
    loop {
        let ready = match data.epoll.wait(&mut events, 15000) {
            Ok(v) => v,
            Err(nix::errno::Errno::EINTR) => 0,
            Err(e) => panic!("Epoll wait failed: {:?}", e)
        };
        for i in 0..ready {
            let event = &events[i];
            match event.data() {
                EPOLL_SIGFD_ID => {
                    let sig = data.sigfd.read_signal().unwrap().unwrap();
                    if sig.ssi_signo == libc::SIGCHLD as u32 {
                        //println!("Got SIGCHLD, ignoring");
                    } else {
                        let s = nix::sys::signal::Signal::try_from(sig.ssi_signo as i32).unwrap();
                        println!("sigfd event, treated as exit command. Signal: {s:#?}");
                        SHUTDOWN_STARTED.store(true, std::sync::atomic::Ordering::Relaxed);
                        for (id, _) in &data.services {
                            data.work_queue.as_ref().unwrap().send(WorkItem::StopService(*id, false)).unwrap();
                        }
                    }
                }
                EPOLL_WEBSERVER_ID => {
                    crate::web::webserver_accept();
                }
                id => {
                    let service = match data.services.get_mut(&id) {
                        None => { panic!("Got epoll event for id '{}' but no service found\n{:#?}", id, data.services); }
                        Some(v) => v,
                    };
                    let name = &service.config.name;
                    let pid = service.pid.take().unwrap();
                    data.epoll.delete(pid.1).expect("Failed to delete from epoll");
                    let status = nix::sys::wait::waitpid(pid.0, None).expect("waitpid failed");
                    match status {
                        nix::sys::wait::WaitStatus::Exited(_, code) => println!("{name} exited with status {code}"),
                        nix::sys::wait::WaitStatus::Signaled(_, sig, _) => println!("{name} exited with signal {sig}"),
                        _ => eprintln!("Unknown status {status:#?}")
                    }
                    service.reaped.store(true, std::sync::atomic::Ordering::Relaxed);
                    data.work_queue.as_ref().unwrap().send(WorkItem::StopService(id, false)).unwrap();
                }
            }
        }
        if SHUTDOWN_STARTED.load(std::sync::atomic::Ordering::Relaxed) && !data.any_running() {
            break;
        }
    }

    Data::get().work_queue.as_ref().unwrap().send(WorkItem::StopService(0, true)).unwrap();

    starter.join().unwrap();
}

fn starter_thread(autostart: Vec<String>) {
    let data = Data::get();
    let (recv, mut todo) = {
        let (s,r) = std::sync::mpsc::channel();
        data.work_queue.replace(s);

        let mut id_map = std::collections::HashMap::new();
        for (_, service) in &data.services {
            id_map.insert(service.config.name.clone(), service.id);
        }

        let todo = autostart.into_iter()
            .map(|s| WorkItem::StartService(*id_map.get(&s).unwrap()))
            .collect::<VecDeque<_>>();

        (r, todo)
    };

    'main_loop: loop {
        while let Some(work) = todo.pop_front() {
            match work {
                WorkItem::StartService(id) => {
                    if SHUTDOWN_STARTED.load(std::sync::atomic::Ordering::Relaxed) { continue; }
                    let (missing_deps, dep_fail) = {
                        let mut missing_deps = Vec::new();
                        let mut dep_fail = false;
                        let service = data.services.get(&id).unwrap();
                        for dep in &service.dependencies {
                            let dep = data.services.get(dep).unwrap();
                            if !dep.started() {
                                if dep.can_autostart() {
                                    missing_deps.push(WorkItem::StartService(dep.id));
                                } else {
                                    dep_fail = true;
                                    break;
                                }
                            }
                        }
                        (missing_deps, dep_fail)
                    };
                    let service = data.services.get_mut(&id).unwrap();
                    if dep_fail {
                        service.status = ServiceStatus::FAILED("Failed to start dependency".into());
                        service.restart_count = 3;
                        continue;
                    } else if !missing_deps.is_empty() {
                        for dep in missing_deps { todo.push_front(dep); }
                        todo.push_back(WorkItem::StartService(id));
                        continue;
                    }
                    service.start();
                }
                WorkItem::StopService(id, manual) => {
                    let can_stop = {
                        let mut can_stop = true;
                        let service = match data.services.get(&id) {
                            Some(v) => v,
                            None => continue
                        };
                        for dep in &service.dependents {
                            let dep = data.services.get(dep).unwrap();
                            if !dep.stopped() {
                                todo.push_front(WorkItem::StopService(dep.id, manual));
                                can_stop = false;
                            }
                        }
                        can_stop
                    };
                    let service = data.services.get_mut(&id).unwrap();
                    if !can_stop {
                        todo.push_back(WorkItem::StopService(id, manual));
                        continue;
                    }
                    service.stop(manual);
                }
            }
        }

        let work = recv.recv().unwrap();
        match work { WorkItem::StopService(0, _) => break 'main_loop, _ => {} }
        todo.push_back(work);
    }
}